package TK23_2023.module.complex;

import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageBuilding;
import adf.core.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.core.agent.communication.standard.bundle.information.MessageRoad;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.algorithm.PathPlanning;
import adf.core.component.module.complex.RoadDetector;
import adf.core.debug.DefaultLogger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.apache.log4j.Logger;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class TkRoadDetector extends RoadDetector {

  private Set<Area> openedAreas = new HashSet<>();
  //private Clustering clustering;
  private PathPlanning pathPlanning;

  private EntityID result;
  private Logger logger;



  ArrayList<StandardEntity> policeList;
  ArrayList<StandardEntity> fireList;
  ArrayList<StandardEntity> refugeList;
  int localId = 0;
  StandardEntity responder = null;
  StandardEntity myRefuge = null;

  public TkRoadDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    logger = DefaultLogger.getLogger(agentInfo.me());
    this.pathPlanning = moduleManager.getModule("TkRoadDetector.PathPlanning", "adf.impl.module.algorithm.DijkstraPathPlanning");
    //this.clustering = moduleManager.getModule("SampleRoadDetector.Clustering", "adf.impl.module.algorithm.KMeansClustring");
    //registerModule(this.clustering);
    registerModule(this.pathPlanning);
    this.result = null;

    policeList = new ArrayList<>(this.worldInfo.getEntitiesOfType(POLICE_FORCE));
    fireList = new ArrayList<>(this.worldInfo.getEntitiesOfType(FIRE_BRIGADE));
    refugeList = new ArrayList<>(this.worldInfo.getEntitiesOfType(REFUGE));

    localId = policeList.indexOf(agentInfo.me());
    //System.out.println(localId);

    int skip = localId;
    while (fireList.size() > 0 && responder == null) {
      for (StandardEntity e : fireList) {
        if (skip-- > 0) { continue; }
        responder = e;
        break;
      }
    }

    skip = localId;
    while (refugeList.size() > 0 && myRefuge == null) { //TODO: Refactor
      for (StandardEntity e : refugeList) {
        if (skip-- > 0) { continue; }
        myRefuge = e;
        break;
      }
    }
  }


  @Override
  public RoadDetector updateInfo(MessageManager messageManager) {
    logger.debug("Time:" + agentInfo.getTime());
    super.updateInfo(messageManager);

    for (CommunicationMessage message : messageManager.getReceivedMessageList(MessagePoliceForce.class)) {
      MessagePoliceForce messagePoliceForce = (MessagePoliceForce) message;
      StandardEntity entity = worldInfo.getEntity(messagePoliceForce.getPosition());
      if (entity.getStandardURN() == BUILDING || entity.getStandardURN() == REFUGE) {
        openedAreas.add((Area) entity);
        //System.out.println(entity.getID().getValue() + " is clear");
      }
    }

    for (CommunicationMessage message : messageManager.getReceivedMessageList(MessageRoad.class)) {
      MessageRoad messageRoad = (MessageRoad) message;
      if (messageRoad.getBlockadeID() == null) {
        StandardEntity entity = worldInfo.getEntity(messageRoad.getRoadID());
        openedAreas.add((Area) entity);
        //System.out.println(entity.getID().getValue() + " is clear");
      }
    }

    return this;
  }


  @Override
  public RoadDetector calc() {
    EntityID positionID = this.agentInfo.getPosition();
    StandardEntity currentPosition = worldInfo.getEntity(positionID);

    if (currentPosition.getStandardURN().equals(BUILDING)) {
      openedAreas.add((Area) currentPosition);
    }

    //openedAreas.add((Area) currentPosition);
    switch (currentPosition.getStandardURN()) {
      case ROAD -> {
        Road road = (Road) currentPosition;
        if (!road.isBlockadesDefined() || road.getBlockades().size() <= 0) {
          openedAreas.add((Area) currentPosition);
        }
        break;
      }
      default -> {
        openedAreas.add((Area) currentPosition);
      }
    }

    if (positionID.equals(result)) {
      logger.debug("reach to " + currentPosition + " resetting target");
      this.result = null;
    }

    if (openedAreas.contains(worldInfo.getEntity(result))) {
      this.result = null;
    }

    if (this.result == null) {
      HashSet<Area> currentTargets = calcTargets();
      logger.debug("Targets: " + currentTargets);
      if (currentTargets.isEmpty()) {
        this.result = null;
        return this;
      }
      this.pathPlanning.setFrom(positionID);
      this.pathPlanning.setDestination(toEntityIds(currentTargets));
      List<EntityID> path = this.pathPlanning.calc().getResult();
      if (path != null && path.size() > 0) {
        this.result = path.get(path.size() - 1);
      }
      logger.debug("Selected Target: " + this.result);
    }

    return this;
  }


  private Collection<EntityID> toEntityIds(Collection<? extends StandardEntity> entities) {
    ArrayList<EntityID> eids = new ArrayList<>();
    for (StandardEntity standardEntity : entities) {
      eids.add(standardEntity.getID());
    }
    return eids;
  }


  private HashSet<Area> calcTargets() {
    HashSet<Area> targetAreas = new HashSet<>();

    if (isValidHuman(responder)) {
      Human h = (Human) responder;
      Area a = (Area) worldInfo.getEntity(h.getPosition());
      targetAreas.add(a);
      for (EntityID id : a.getNeighbours()) {
        try {
          targetAreas.add((Area) worldInfo.getEntity(id));
        } catch (Exception e) {
          //NOP
        }
      }
    }

    if (agentInfo.getTime() <= 10) {
      targetAreas.add(agentInfo.getPositionArea());
      targetAreas.addAll(agentInfo.getPositionArea().getNeighbours().stream().map(id->(Area)worldInfo.getEntity(id)).toList());
    }

    targetAreas.add((Area) myRefuge);
    targetAreas.removeAll(openedAreas);
    //for (StandardEntity e : this.worldInfo.getEntitiesOfType(REFUGE, GAS_STATION)) {
    /*
    for (StandardEntity e : this.worldInfo.getEntitiesOfType(REFUGE)) {
      targetAreas.add((Area) e);
    }
     */
    if (targetAreas.size() <= 0) {
      //for (StandardEntity e : this.worldInfo.getEntitiesOfType(CIVILIAN, AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE)) {
      for (StandardEntity e : this.worldInfo.getEntitiesOfType(CIVILIAN)) {
        if (isValidHuman(e)) {
          Human h = (Human) e;
          targetAreas.add((Area) worldInfo.getEntity(h.getPosition()));
        }
      }
      //HashSet<Area> inClusterTarget = filterInCluster(targetAreas);
      //inClusterTarget.removeAll(openedAreas);
      //return inClusterTarget;
      targetAreas.removeAll(openedAreas);
    }
    return targetAreas;
  }


  /*
  private HashSet<Area> filterInCluster(HashSet<Area> targetAreas) {
    int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
    HashSet<Area> clusterTargets = new HashSet<>();
    HashSet<StandardEntity> inCluster = new HashSet<>(
        clustering.getClusterEntities(clusterIndex));
    for (Area target : targetAreas) {
      if (inCluster.contains(target))
        clusterTargets.add(target);

    }
    return clusterTargets;
  }
   */


  @Override
  public EntityID getTarget() {
    return this.result;
  }


  private boolean isValidHuman(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    if (!target.isHPDefined() || target.getHP() == 0)
      return false;
    if (!target.isPositionDefined())
      return false;
    if (!target.isDamageDefined() || target.getDamage() == 0)
      return false;
    if (!target.isBuriednessDefined())
      return false;

    StandardEntity position = worldInfo.getPosition(target);
    if (position == null)
      return false;

    StandardEntityURN positionURN = position.getStandardURN();
    if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM)
      return false;

    return true;
  }
}