package TK23_2023.module.complex;

import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.core.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.complex.HumanDetector;
import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class TkTransportingHumanDetector extends HumanDetector {

  //private Clustering clustering;

  private EntityID result;

  private Logger logger;



  HashSet<EntityID> targetSet = new HashSet<>();
  HashSet<EntityID> subTargetSet = new HashSet<>();
  ArrayList<StandardEntity> ambulanceList;
  int localId = 0;



  public TkTransportingHumanDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    logger = DefaultLogger.getLogger(agentInfo.me());
    /*
    this.clustering = moduleManager.getModule("SampleHumanDetector.Clustering",
        "adf.impl.module.algorithm.KMeansClustering");
    registerModule(this.clustering);

     */
    ambulanceList = new ArrayList<>(this.worldInfo.getEntitiesOfType(AMBULANCE_TEAM));
    localId = ambulanceList.indexOf(agentInfo.me());
  }


  @Override
  public HumanDetector updateInfo(MessageManager messageManager) {
    logger.debug("Time:" + agentInfo.getTime());
    super.updateInfo(messageManager);


    for (CommunicationMessage message : messageManager.getReceivedMessageList(MessageAmbulanceTeam.class)) {
      MessageAmbulanceTeam messageAmbulanceTeam = (MessageAmbulanceTeam) message;
      if (messageAmbulanceTeam.getAgentID().equals(agentInfo.getID())) { continue; }
      StandardEntity entity = worldInfo.getEntity(messageAmbulanceTeam.getTargetID());
      if (entity.getStandardURN() == CIVILIAN) {
        result = null;
        //System.out.println(entity.getID().getValue() + " is clear");
      }
    }

    return this;
  }


  @Override
  public HumanDetector calc() {
    Human transportHuman = this.agentInfo.someoneOnBoard();
    if (transportHuman != null) {
      logger.debug("someoneOnBoard:" + transportHuman);
      this.result = transportHuman.getID();
      return this;
    }

    targetSet = new HashSet<>(filterTransportTargets(targetSet.stream().map(id -> worldInfo.getEntity(id)).toList()).stream().map(e->e.getID()).toList());
    targetSet.addAll(filterTransportTargets(this.worldInfo.getEntitiesOfType(CIVILIAN)).stream().map(e->e.getID()).toList());
    subTargetSet = new HashSet<>(filterRescueTargets(targetSet.stream().map(id -> worldInfo.getEntity(id)).toList()).stream().map(e->e.getID()).toList());
    subTargetSet.addAll(filterRescueTargets(this.worldInfo.getEntitiesOfType(CIVILIAN)).stream().map(e->e.getID()).toList());

    if (this.result != null) {
      Human target = (Human) this.worldInfo.getEntity(this.result);
      if (!isValidHuman(target) || !targetSet.contains(result)) {
        logger.debug("Invalid Human:" + target + " ==>reset target");
        this.result = null;
      }
    }
    if (this.result == null) {
      this.result = calcTarget();
    }
    return this;
  }


  private EntityID calcTarget() {
    //List<Human> rescueTargets = filterTransportTargets( this.worldInfo.getEntitiesOfType(CIVILIAN));
    //List<Human> rescueTargetsInCluster = filterInCluster(rescueTargets);
    ArrayList<Human> targets = new ArrayList<>(targetSet.stream().map(id->(Human)worldInfo.getEntity(id)).toList());
    /*
    List<Human> targets = rescueTargetsInCluster;
    if (targets.isEmpty())
      targets = rescueTargets;

     */

    logger.debug("Targets:" + targets);
    if (!targets.isEmpty()) {
      //targets.sort(new DistanceSorter(this.worldInfo, this.agentInfo.me()));
      Human selected = targets.get(localId % targets.size());
      if (localId < targets.size()) {
        logger.debug("Selected:" + selected);
        return selected.getID();
      }
    }


    targets.addAll(subTargetSet.stream().map(id->(Human)worldInfo.getEntity(id)).toList());
    if (!targets.isEmpty()) {
      Human selected = targets.get(localId % targets.size());
      return selected.getID();
    }

    return null;
  }


  @Override
  public EntityID getTarget() {
    return this.result;
  }


  private List<Human> filterTransportTargets(Collection<? extends StandardEntity> list) {
    List<Human> rescueTargets = new ArrayList<>();
    for (StandardEntity next : list) {
      if (!(next instanceof Human))
        continue;
      Human h = (Human) next;
      if (!isValidHuman(h))
        continue;
      if (h.getBuriedness() == 0 && h.getDamage() > 0) {
        rescueTargets.add(h);
      }
    }
    return rescueTargets;
  }

  private List<Human> filterRescueTargets(Collection<? extends StandardEntity> list) {
    List<Human> rescueTargets = new ArrayList<>();
    for (StandardEntity next : list) {
      if (!(next instanceof Human))
        continue;
      Human h = (Human) next;
      if (!isValidHuman(h))
        continue;
      if (h.getBuriedness() == 0)
        continue;
      rescueTargets.add(h);
    }
    return rescueTargets;
  }


  /*
  private List<Human>
      filterInCluster(Collection<? extends StandardEntity> entities) {
    int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
    List<Human> filter = new ArrayList<>();
    HashSet<StandardEntity> inCluster = new HashSet<>(
        clustering.getClusterEntities(clusterIndex));
    for (StandardEntity next : entities) {
      if (!(next instanceof Human))
        continue;
      Human h = (Human) next;
      if (!h.isPositionDefined())
        continue;
      StandardEntity position = this.worldInfo.getPosition(h);
      if (position == null)
        continue;
      if (!inCluster.contains(position))
        continue;
      filter.add(h);
    }
    return filter;

  }

   */

  private class DistanceSorter implements Comparator<StandardEntity> {

    private StandardEntity reference;
    private WorldInfo worldInfo;

    DistanceSorter(WorldInfo wi, StandardEntity reference) {
      this.reference = reference;
      this.worldInfo = wi;
    }


    public int compare(StandardEntity a, StandardEntity b) {
      int d1 = this.worldInfo.getDistance(this.reference, a);
      int d2 = this.worldInfo.getDistance(this.reference, b);
      return d1 - d2;
    }
  }

  private boolean isValidHuman(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    if (!target.isHPDefined() || target.getHP() == 0)
      return false;
    if (!target.isPositionDefined())
      return false;
    if (!target.isDamageDefined() || target.getDamage() == 0)
      return false;
    if (!target.isBuriednessDefined())
      return false;

    StandardEntity position = worldInfo.getPosition(target);
    if (position == null)
      return false;

    StandardEntityURN positionURN = position.getStandardURN();
    if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM)
      return false;

    return true;
  }
}